/*
 * Copyright (c) 2007 BUSINESS OBJECTS SOFTWARE LIMITED
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 *     * Neither the name of Business Objects nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


/*
 * FeatureName.java
 * Creation date: Dec 9, 2004.
 * By: Edward Lam
 */
package org.openquark.cal.services;



/**
 * A class that can be used to uniquely identify a feature.
 * @author Edward Lam
 */
public abstract class FeatureName {
    
    /**
     * A class for the feature type enum pattern that identifies the type of a CAL feature.
     * @author Frank Worsley
     */
    public static final class FeatureType {
        
        private final String typeName;
        
        FeatureType(String typeName) {
            
            if (typeName == null) {
                throw new NullPointerException();
            }
            
            this.typeName = typeName;
        }
        
        public String getTypeName() {
            return typeName;
        }
        
        @Override
        public String toString() {
            return getTypeName();
        }
    }

    /** The type of this feature. */
    private final FeatureType type;
    
    /** The name of this feature. */
    private final String name;
    
    /**
     * Constructs a new FeatureName.
     * @param type the type of the feature the name is for
     * @param name the name of the feature
     */
    FeatureName(FeatureType type, String name) {
        
        if (type == null || name == null) {
            throw new NullPointerException();
        }
        
        this.type = type;
        this.name = name;
    }
    
    /**
     * @return the type of the feature the name is for
     */
    public FeatureType getType() {
        return type;
    }
    
    /**
     * @return the string form of the name of the feature
     */
    public String getName() {
        return name;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return type + ": " + name;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object o) {
        
        if (o.getClass().equals(getClass())) {
            return ((FeatureName) o).getType() == type &&
                   ((FeatureName) o).getName().equals(name);
        }
        
        return false;
    }
    
    /**
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        return getClass().hashCode() * 17 + toString().hashCode();
    }
    

}
